/**
 * @vue-async/module-loader@1.0.7
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@vue-async/utils'), require('vue')) :
  typeof define === 'function' && define.amd ? define(['exports', '@vue-async/utils', 'vue'], factory) :
  (global = global || self, factory(global.VueAsyncModuleLoader = {}, global.VueAsyncUtils, global.Vue));
}(this, (function (exports, utils, Vue) { 'use strict';

  Vue = Vue && Object.prototype.hasOwnProperty.call(Vue, 'default') ? Vue['default'] : Vue;

  const isProduction = "development" === 'production';
  const namespaces = 'dynamicComponent';
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  function dynamicComponent (Vue, store) {
      return {
          /**
           * add component to store
           * @param component vue component
           * @param position position
           */
          add(component, position) {
              if (!(typeof component === 'object' || typeof component === 'function')) {
                  return utils.error(isProduction, '[moduleLoader] dynamicComponent function "add" only accept object or function component as param.');
              }
              store.commit(`${namespaces}/add`, {
                  component,
                  position,
              });
          },
          /**
           * remove component from store
           * @param name component name
           * @param position position
           */
          remove(name, position) {
              if (typeof name !== 'string') {
                  return utils.error(isProduction, '[moduleLoader] dynamicComponent function "remove" only accept string name of component as param.');
              }
              store.commit(`${namespaces}/remove`, {
                  name,
                  position,
              });
          },
      };
  }

  const isProduction$1 = "development" === 'production';
  var dynamicComponentState = {
      namespaced: true,
      state: {
          GLOBAL: {},
      },
      mutations: {
          // 注册插槽组件
          add(state, { component, position = 'GLOBAL' }) {
              const _component = component.component || component;
              // 必须使用外部 Vue 对象，本地 import 对象不一致
              const componentName = component.name || utils.getComponentName(_component);
              if (!componentName) {
                  return utils.error(isProduction$1, '[moduleLoader] component must be named for dynamic component.');
              }
              if (!state[position]) {
                  state[position] = {};
              }
              if (state[position][componentName]) {
                  return utils.error(isProduction$1, `[moduleLoader] async component "${componentName}" has existed in the ${position}, ignored it.`);
              }
              state[position][componentName] = component;
          },
          // 移除注册插槽组件
          remove(state, { name, position = 'GLOBAL' }) {
              if (!state[position]) {
                  return utils.error(isProduction$1, `[moduleLoader] position "${position}" does not  exist.`);
              }
              if (!state[position][name]) {
                  return utils.error(isProduction$1, `[moduleLoader] async component "${name}" does not  exist in the ${position}.`);
              }
              delete state[position][name];
          },
      },
  };

  const isProduction$2 = "development" === 'production';
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  function createEventBus (Vue) {
      let events = {};
      return {
          /**
           * 发射事件
           * @param {string} eventName 事件名称
           * @param {any} payload 事件载荷
           */
          emit(eventName, payload) {
              if (!events[eventName]) {
                  return utils.warn(isProduction$2, `[moduleLoader] event "${eventName}" is triggered, but no listeners，nothing will be happening.`);
              }
              for (const func of events[eventName].values()) {
                  func(payload);
              }
          },
          /**
           * 事件监听
           * @param eventName 事件名称
           * @param handler 处理事件的方法
           */
          on(eventName, handler) {
              if (typeof handler === 'function') {
                  if (!events[eventName])
                      events[eventName] = new Set();
                  events[eventName].add(handler);
              }
              else {
                  utils.error(isProduction$2, `[moduleLoader] handler must be a function, but recived ${typeof handler}`);
              }
          },
          /**
           * 取消事件监听
           * @param eventName 事件名称
           * @param handler 处理事件的方法
           */
          off(eventName, handler) {
              if (!handler) {
                  return utils.warn(false, '[moduleLoader] handler must be a valid event function');
              }
              else if (!events.hasOwnProperty(eventName)) {
                  return utils.error(isProduction$2, `[moduleLoader] no event name of "${eventName}" is on listening`);
              }
              events[eventName].delete(handler);
          },
          /**
           * 清理事件总线
           */
          clear() {
              events = {};
          },
          /**
           * 获取当前事件总线详情
           */
          getEvents() {
              return events;
          },
      };
  }

  /**
   * @author Hubert
   * fork from https://github.com/kuitos/import-html-entry/blob/master/src/utils.js
   */
  const isIE11 = typeof navigator !== 'undefined' && navigator.userAgent.indexOf('Trident') !== -1;
  function shouldSkipProperty(global, p) {
      if (!global.hasOwnProperty(p) || (!isNaN(p) && parseInt(p) < global.length))
          return true;
      if (isIE11) {
          // https://github.com/kuitos/import-html-entry/pull/32，最小化 try 范围
          try {
              return global[p] && global[p].parent === window;
          }
          catch (err) {
              return true;
          }
      }
      else {
          return false;
      }
  }
  // safari unpredictably lists some new globals first or second in object order
  let firstGlobalProp, secondGlobalProp, lastGlobalProp;
  function getGlobalProp(global) {
      let cnt = 0;
      let lastProp;
      let hasIframe = false;
      for (const p in global) {
          if (shouldSkipProperty(global, p))
              continue;
          // 遍历 iframe，检查 window 上的属性值是否是 iframe，是则跳过后面的 first 和 second 判断
          for (let i = 0; i < window.frames.length && !hasIframe; i++) {
              const frame = window.frames[i];
              if (frame === global[p]) {
                  hasIframe = true;
                  break;
              }
          }
          if (!hasIframe && ((cnt === 0 && p !== firstGlobalProp) || (cnt === 1 && p !== secondGlobalProp)))
              return p;
          cnt++;
          lastProp = p;
      }
      if (lastProp !== lastGlobalProp)
          return lastProp;
      return;
  }
  function noteGlobalProps(global) {
      // alternatively Object.keys(global).pop()
      // but this may be faster (pending benchmarks)
      firstGlobalProp = secondGlobalProp = undefined;
      for (const p in global) {
          if (shouldSkipProperty(global, p))
              continue;
          if (!firstGlobalProp)
              firstGlobalProp = p;
          else if (!secondGlobalProp)
              secondGlobalProp = p;
          lastGlobalProp = p;
      }
      return lastGlobalProp;
  }
  /** 加载 entry 脚本 */
  function execScript(entry, proxy = window) {
      return new Promise((resolve, reject) => {
          noteGlobalProps(proxy);
          const script = document.createElement('script');
          script.src = entry;
          script.onload = () => {
              const propName = getGlobalProp(proxy);
              const exports = propName ? proxy[propName] || {} : {};
              resolve(exports);
          };
          script.onerror = () => {
              utils.error("development" === 'production', `[moduleLoader] script had a problem to create, entry：${entry}`);
              reject(new Error('script load error'));
          };
          document.body.appendChild(script);
      });
  }
  /** 加载 styles */
  function execStyles(styles, styleFor) {
      // load css
      if (styles.length) {
          styles.map((href) => {
              const link = document.createElement('link');
              link.rel = 'stylesheet';
              link.type = 'text/css';
              link.href = href;
              styleFor && (link.dataset.styleFor = styleFor);
              document.getElementsByTagName('head')[0].appendChild(link);
          });
      }
  }

  const isProduction$3 = "development" === 'production';
  /** 格式化模块配置 */
  function formatModules(modules, errorCallback = () => { }) {
      if (Array.isArray(modules)) {
          return modules.reduce((prev, curr) => [...prev, ...formatModules(curr)], []);
      }
      else if (utils.isPlainObject(modules)) {
          const _modules = modules;
          // no lib name
          if (_modules.entry && typeof _modules.entry === 'string') {
              return [
                  {
                      ...modules,
                      styles: _modules.styles ? (!Array.isArray(_modules.styles) ? [_modules.styles] : _modules.styles) : [],
                  },
              ];
          }
          else {
              return Object.keys(_modules).map((moduleName) => typeof _modules[moduleName] === 'string'
                  ? {
                      entry: _modules[moduleName],
                      moduleName,
                  }
                  : {
                      ..._modules[moduleName],
                      styles: _modules[moduleName].styles
                          ? !Array.isArray(_modules[moduleName].styles)
                              ? [_modules[moduleName].styles]
                              : _modules[moduleName].styles
                          : [],
                      moduleName,
                  });
          }
      }
      else if (utils.isFunction(modules)) {
          return [modules];
      }
      else {
          errorCallback(`not support "${typeof modules}" type module`, modules);
          return [];
      }
  }
  /** 校验子应用导出的 生命周期 对象是否正确 */
  function validateExportLifecycle(exports) {
      // export default function(){} 兼容处理
      const _export = (exports && exports.default) || exports;
      if (utils.isFunction(_export)) {
          return true;
      }
      else {
          // 生命周期
          // todo: 延迟加载
          // const { bootstrap ，mount, unmount} = exports ?? {};
          // return isFunction(bootstrap) && isFunction(mount) && isFunction(unmount);
          const { bootstrap } = exports || {};
          return utils.isFunction(bootstrap);
      }
  }
  /** 获取生命周期对象 */
  function getLifecyclesFromExports(scriptExports, moduleName, global) {
      if (validateExportLifecycle(scriptExports)) {
          const _export = (scriptExports && scriptExports.default) || scriptExports;
          return utils.isFunction(_export) ? { bootstrap: _export } : scriptExports;
      }
      utils.warn(isProduction$3, `[moduleLoader] lifecycle not found from ${moduleName} entry exports, fallback to get from window['${moduleName}']`);
      // fallback to global variable who named with ${moduleName} while module exports not found
      const globalVariableExports = global[moduleName];
      if (validateExportLifecycle(globalVariableExports)) {
          const _export = (globalVariableExports && globalVariableExports.default) || globalVariableExports;
          return utils.isFunction(_export) ? { bootstrap: _export } : globalVariableExports;
      }
      throw new Error(`[moduleLoader] You need to export lifecycle functions in ${moduleName} entry`);
  }
  var createModuleLoader = (Vue, status) => {
      return function loader(modules, opts = {}) {
          const { success, error = (msg) => {
              utils.error(isProduction$3, msg);
          }, } = opts;
          status.current = false;
          // eslint-disable-next-line @typescript-eslint/no-this-alias
          const _self = this;
          const _modules = formatModules(modules, error);
          // todo: load in sandbox
          const global = window;
          return new Promise((resolve) => Promise.all(_modules.map((module) => exec(module)))
              .then(() => {
              status.current = true;
          })
              .then(success || resolve));
          function exec(module) {
              // local module
              if (utils.isFunction(module)) {
                  const result = module.call(_self, Vue);
                  if (result && result instanceof Promise) {
                      return result;
                  }
                  return Promise.resolve();
              }
              // remote module
              else {
                  // server render
                  if (Vue.prototype.$isServer) {
                      return Promise.resolve();
                  }
                  const { moduleName, entry, styles = [], args } = module;
                  // load styles
                  execStyles(styles, moduleName);
                  // exec script
                  return execScript(entry, global)
                      .then((scriptExports) => {
                      const { bootstrap } = getLifecyclesFromExports(scriptExports, moduleName, global);
                      bootstrap.call(_self, Vue, args);
                  })
                      .catch((err) => {
                      // 异常不阻止当前执行，error 中处理
                      try {
                          error(err.message, module);
                      }
                      catch { }
                  });
              }
          }
      };
  };

  /**
   * componentLoader
   */
  const isProduction$4 = "development" === 'production';
  /** 验证组件导出是否正确 */
  function validateExportComponent(exports) {
      const _export = (exports && exports.default) || exports;
      if (utils.isPlainObject(_export)) {
          return (Object.keys(_export).length > 0 &&
              (utils.hasOwn(_export, 'template') ||
                  utils.hasOwn(_export, 'render') || // component definition
                  utils.hasOwn(_export, 'component') || // AsyncComponentFactory
                  utils.isFunction(_export)) // AsyncComponentPromise
          );
      }
      else {
          return _export instanceof Vue;
      }
  }
  /** 获取组件 */
  function getComponentFromExport(scriptExports, componentName, global) {
      if (validateExportComponent(scriptExports)) {
          return (scriptExports && scriptExports.default) || scriptExports;
      }
      utils.warn(isProduction$4, `[moduleLoader] component not found from ${componentName} entry exports, fallback to get from window['${componentName}']`);
      // fallback to global variable who named with ${moduleName} while module exports not found
      const globalVariableExports = global[componentName];
      if (validateExportComponent(globalVariableExports)) {
          return (globalVariableExports && globalVariableExports.default) || globalVariableExports;
      }
      throw new Error(`[moduleLoader] You need to export vue component in ${componentName} entry`);
  }
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  var createComponentLoader = (Vue) => {
      return function loader(componentName, src) {
          const global = window;
          return execScript(src, global).then((scriptExports) => {
              return getComponentFromExport(scriptExports, componentName, global);
          });
      };
  };

  function install(Vue) {
      if (install.installed)
          return;
      install.installed = true;
      // 用于监听模块加载的情况，处理路由404问题
      const vm = new Vue({
          data() {
              return {
                  status: {
                      current: false,
                  },
              };
          },
      });
      Object.defineProperties(Vue.prototype, {
          $eventBus: {
              value: createEventBus(),
              writable: false,
          },
          $moduleLoader: {
              value: createModuleLoader(Vue, vm.status),
              writable: false,
          },
          $componentLoader: {
              value: createComponentLoader(),
              writable: false,
          },
      });
      /**
       * router注入（当路由被注册注入时，必须调用$moduleLoaser至少一次，否则 status 状态无法变更造成路由被阻止）
       * @param router
       */
      const routerInject = (router) => {
          return router.beforeEach((to, from, next) => {
              if (!to.name || to.name === '404' || to.name === 'page-not-found' || to.name === 'not-found' || to.path === '*') {
                  // 模块已经被加载完成, 但由于在其之前添加 beforeEach 阻止时间过长，vm.$watch还没开始监听
                  if (vm.status.current) {
                      return resolveRoute(to, from, next);
                  }
                  // 当 status 为 false 时watch状态
                  vm.$watch(() => vm.status.current, (newVal, oldVal) => {
                      // 从false => true状态
                      if (newVal && !oldVal) {
                          resolveRoute(to, from, next);
                      }
                  });
              }
              else {
                  next();
              }
          });
          // 解决动态路由404问题
          function resolveRoute(to, from, next) {
              const fullPath = to.redirectedFrom || to.fullPath;
              const { resolved, location } = router.resolve(fullPath);
              // 在加载完组件后resolve的地址与原来要跳转的地址不一致时跳转
              // 以免造成404死循环
              if (resolved.name !== to.name) {
                  next(location);
              }
              else {
                  next();
              }
          }
      };
      /**
       * store 注入
       * 添加 $dynamicComponent 方法
       */
      const storeInject = (store) => {
          store.registerModule(namespaces, dynamicComponentState);
          // define $dynamicComponent
          Object.defineProperty(Vue.prototype, '$dynamicComponent', {
              value: dynamicComponent(Vue, store),
              writable: false,
          });
      };
      const _init = Vue.prototype._init;
      /**
       * 从 Vue root option 中获取 router | store 实例
       */
      Vue.prototype._init = function (options = {}) {
          // router beforeEach 需要在 beforeCreate 之前添加，才能在页面强制刷新时第一次生效
          if (options.router) {
              routerInject(options.router);
          }
          if (options.store) {
              storeInject(options.store);
          }
          _init.call(this, options);
      };
      // Used to avoid multiple mixins being setup
      // when in dev mode and hot module reload
      // https://github.com/vuejs/vue/issues/5089#issuecomment-284260111
      if (Vue.$__module_loader_installed__)
          return;
      // eslint-disable-next-line @typescript-eslint/camelcase
      Vue.$__module_loader_installed__ = true;
      if (window && !window.Vue) {
          window.Vue = Vue;
      }
      Vue.mixin({
          beforeCreate() {
              const options = this.$options;
              if (options.moduleLoader) {
                  options.moduleLoader.init(this, options.ssrContext);
                  this.$moduleLoadManager = Vue.observable(options.moduleLoader.framework);
              }
              else {
                  this.$moduleLoadManager = (options.parent && options.parent.$moduleLoadManager) || this;
              }
          },
      });
  }

  class ModuleLoader {
      constructor(options) {
          this._framework = {
              layouts: {},
          };
          this._options = options;
      }
      get framework() {
          return this._framework;
      }
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      init(root, ssrContext) {
          this._options &&
              Object.entries(this._options).map(([key, value]) => {
                  if (!utils.hasOwn(this._framework, key)) {
                      this._framework[key] = value;
                  }
                  else {
                      utils.warn("development" === 'production', `参数 ${key} 已在在`);
                  }
              });
          // 默认方法
          !utils.hasOwn(this._framework, 'addRoutes') && (this._framework.addRoutes = this._createAddRoutes(root));
          !utils.hasOwn(this._framework, 'addLayouts') && (this._framework.addLayouts = this._createAddLayouts(root));
      }
      _createAddRoutes(root) {
          return (routes) => {
              if (root.$router) {
                  root.$router.addRoutes(routes);
              }
              else {
                  utils.error("development" === 'production', 'vm.$router未定义');
              }
          };
      }
      _createAddLayouts(root) {
          return (key, layout) => {
              if (typeof key === 'string') {
                  root.$set(this._framework.layouts, key, layout);
              }
              else if (utils.isPlainObject(key)) {
                  // plain object
                  const layouts = key;
                  this._framework.layouts = Object.assign(this._framework.layouts, layouts);
              }
              else {
                  utils.error("development" === 'production', 'addLayouts中支持字符串(key)或 Object(key/value)格式');
              }
          };
      }
  }
  ModuleLoader.install = install;
  ModuleLoader.installed = false;
  ModuleLoader.version = '1.0.7';

  const install$1 = ModuleLoader.install;
  ModuleLoader.install = (Vue) => {
      install$1.call(ModuleLoader, Vue);
  };
  // Auto install if it is not done yet and `window` has `Vue`.
  // To allow users to avoid auto-installation in some cases,
  if (window && window.Vue) {
      window.Vue.use(ModuleLoader);
  }

  exports.default = ModuleLoader;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
